@@ -80,6 +80,10 @@ public:
 #if LLVM_VERSION_MAJOR >= 8
                                  ,
                                  SourceLocation openParLoc
+#endif
+#if LLVM_VERSION_MAJOR >= 14
+                                 ,
+                                 bool braced
 #endif
                                  ) override {
     ls_sighelp.activeParameter = (int)currentArg;
@@ -93,7 +97,11 @@ public:
           cand = OverloadCandidate(pattern);
 
       const auto *ccs =
+#if LLVM_VERSION_MAJOR >= 14
+          cand.CreateSignatureString(currentArg, s, *alloc, cCTUInfo, true, braced);
+#else
           cand.CreateSignatureString(currentArg, s, *alloc, cCTUInfo, true);
+#endif
 
       const char *ret_type = nullptr;
       SignatureInformation &ls_sig = ls_sighelp.signatures.emplace_back();
